"""Test classmethods."""

from pathlib import Path
from types import MappingProxyType

from treestamps.config import CommonConfig
from treestamps.grove import Grovestamps, GrovestampsConfig
from treestamps.tree import Treestamps

__all__ = ()


class TestClassMethodds:
    """Test classmethods."""

    def test_get_filename(self):
        """Test get filename."""
        assert Treestamps._get_filename("foo") == ".foo_treestamps.yaml"

    def test_get_wal_filename(self):
        """Test get wal filename."""
        assert Treestamps._get_wal_filename("foo") == ".foo_treestamps.wal.yaml"

    def test_normalize_config(self):
        """Test normalize config."""
        keys = frozenset(["a", "b"])
        cc = CommonConfig("Dummy", program_config_keys=keys)
        assert cc.program_config is None

        cc = CommonConfig("Dummy", program_config_keys=keys, program_config={})
        assert cc.program_config == MappingProxyType({})

        cc = CommonConfig("Dummy", program_config_keys=keys, program_config={"a": 1})
        assert cc.program_config == MappingProxyType({"a": 1})

        cc = CommonConfig(
            "Dummy",
            program_config_keys=keys,
            program_config={"a": 1, "b": [3, 2, 1, 2]},
        )
        assert cc.program_config == MappingProxyType(
            {
                "a": 1,
                "b": (1, 2, 2, 3),
            }
        )

        cc = CommonConfig(
            "Dummy", program_config_keys=keys, program_config={"a": {"b": [2, 1, 3, 1]}}
        )
        assert cc.program_config == MappingProxyType({"a": {"b": (1, 1, 2, 3)}})

    def test_grove(self):
        """Test the factory."""
        path_a = Path(__file__)
        path_b = path_a.parent
        path_c = Path("/tmp")  # noqa: S108

        paths = (path_a, path_b, path_c)

        config = GrovestampsConfig("Dummy", paths=paths)
        cs = Grovestamps(config)

        dirset = {path_b.resolve(), path_c}
        assert set(cs.keys()) == dirset
        dirs = {ts.root_dir for ts in cs.values()}
        assert dirs == dirset



"""Tests init."""

import inspect
from pathlib import Path

PROGRAM = "treestamps"
TEST_FILES_DIR = Path("tests/test_files")
INVALID_DIR = TEST_FILES_DIR / "invalid"
TMP_ROOT = "/tmp"  # noqa:S108


def get_test_dir():
    """Return a module specific tmpdir."""
    frame = inspect.currentframe()
    if frame and frame.f_back:
        caller = frame.f_back
        module_name = caller.f_globals["__name__"]
    else:
        module_name = "unknown"

    return TMP_ROOT / Path(f"{PROGRAM}-{module_name}")



"""Get Methods."""

from pathlib import Path

from treestamps.tree.init import TreestampsInit


class TreestampsGet(TreestampsInit):
    """Get Methods."""

    @staticmethod
    def max_none(a: float | None, b: float | None) -> float | None:
        """None aware max() function."""
        return max((x for x in (a, b) if x is not None), default=None)

    def get(self, path: Path | str) -> float | None:
        """Get the timestamps up the directory tree. All the way to root."""
        mtime: float | None = None
        abs_path = self._get_absolute_path(self.root_dir, path)
        if not abs_path:
            return mtime

        # Walk up the tree to get the maximum time.
        while abs_path != abs_path.parent:
            mtime = self.max_none(mtime, self._timestamps.get(abs_path))
            abs_path = abs_path.parent

        return mtime



"""Base class for testing images."""

import shutil
from pathlib import Path

from tests import TEST_FILES_DIR, get_test_dir


class BaseTestDir:
    """Test images dir."""

    FNS: tuple = ()
    TMP_ROOT: Path = get_test_dir()

    def setup_method(self) -> None:
        """Set up method."""
        self.teardown_method()
        self.TMP_ROOT.mkdir(parents=True)
        for fn in self.FNS:
            src = TEST_FILES_DIR / fn
            dest = self.TMP_ROOT / fn
            shutil.copy(src, dest)

    def teardown_method(self) -> None:
        """Tear down method."""
        shutil.rmtree(self.TMP_ROOT, ignore_errors=True)



"""Dump Methods."""

from termcolor import cprint

from treestamps.tree.write import TreestampsWrite


class TreestampsDump(TreestampsWrite):
    """Dump Methods."""

    def _serialize_timestamps(self):
        """Dumpable timestamp paths need to be strings."""
        dumpable_timestamps = {}
        for abs_path, timestamp in self._timestamps.items():
            try:
                rel_path_str = self._get_relative_path_str(abs_path)
                dumpable_timestamps[rel_path_str] = timestamp
            except Exception as exc:
                cprint(f"WARNING: Serializing {abs_path}: {exc}", "yellow")
        return dumpable_timestamps

    def _cleanup_old_timestamps(self):
        if self._consumed_paths:
            self._consumed_paths.discard(self._dump_path)
            for path in self._consumed_paths:
                try:
                    path.unlink(missing_ok=True)
                except Exception as exc:
                    cprint(f"WARNING: removing old timestamp: {exc}", "yellow")
            self._consumed_paths = set()

    def dump(self) -> None:
        """Serialize timestamps and dump to file."""
        dumpable_timestamps = self._serialize_timestamps()
        self._dump_to_file(self._dump_path, dumpable_timestamps)

        self._cleanup_old_timestamps()



"""Test absolute path resolution."""

from pathlib import Path

from treestamps.tree import Treestamps
from treestamps.tree.config import TreestampsConfig

__all__ = ()


class TestAbsolutePath:
    """Test absolute path resolution."""

    def test_absolute_path_simple(self):
        """Test absolute paths."""
        root = Path.cwd()
        config = TreestampsConfig(program_name="test", path=root)
        treestamps = Treestamps(config=config)
        test_path = root
        abs_path = treestamps._get_absolute_path(test_path, "subdir")
        assert abs_path == test_path / "subdir"

    def test_absolute_path_subpath_absolute(self):
        """Test absolute paths."""
        root = Path("/tmp")  # noqa: S108
        config = TreestampsConfig(program_name="test", path=root)
        treestamps = Treestamps(config=config)
        test_path = root
        abs_path = treestamps._get_absolute_path(test_path, "/tmp/subdir")  # noqa: S108
        assert abs_path == test_path / "subdir"

    def test_absolute_path_subpath_out_of_path(self):
        """Test absolute paths."""
        root = Path.cwd()
        config = TreestampsConfig(program_name="test", path=root, verbose=2)
        treestamps = Treestamps(config=config)
        test_path = root
        abs_path = treestamps._get_absolute_path(test_path, "/tmp")  # noqa: S108
        assert abs_path is None

    def test_absolute_path_subpath_parent(self):
        """Test absolute paths."""
        root = Path("/tmp/subdir/foo")  # noqa: S108
        config = TreestampsConfig(program_name="test", path=root)
        treestamps = Treestamps(config=config)
        test_path = Path("/tmp/subdir")  # noqa: S108
        abs_path = treestamps._get_absolute_path(root, test_path)
        assert abs_path == root

    def test_absolute_path_subpath_dot(self):
        """Test absolute paths."""
        root = Path("/tmp/subdir/foo")  # noqa: S108
        config = TreestampsConfig(program_name="test", path=root)
        treestamps = Treestamps(config=config)
        test_path = "."
        abs_path = treestamps._get_absolute_path(root, test_path)
        assert abs_path == root



"""Common methods."""

from collections.abc import Mapping
from pathlib import Path
from types import MappingProxyType
from typing import TextIO

from ruamel.yaml import YAML
from ruamel.yaml.representer import SafeRepresenter
from termcolor import cprint

from treestamps.tree.config import TreestampsConfig


class TreestampsInit:
    """Common methods."""

    _CONFIG_TAG = "config"
    _TREESTAMPS_CONFIG_TAG = "treestamps_config"
    _WAL_TAG = "wal"
    _FILENAME_TEMPLATE = ".{program_name}_treestamps.yaml"
    _WAL_FILENAME_TEMPLATE = ".{program_name}_treestamps.wal.yaml"

    @staticmethod
    def get_dir(path: Path | str) -> Path:
        """Return a directory for a path."""
        path = Path(path)
        return path if path.is_dir() else path.parent

    @classmethod
    def _get_filename(cls, program_name: str) -> str:
        """Return the timestamps filename for a program."""
        return cls._FILENAME_TEMPLATE.format(program_name=program_name)

    @classmethod
    def _get_wal_filename(cls, program_name: str) -> str:
        """Return the write ahead log filename for the program."""
        return cls._WAL_FILENAME_TEMPLATE.format(program_name=program_name)

    @classmethod
    def get_filenames(cls, program_name):
        """Get all filenames produced by treestamps."""
        return (cls._get_filename(program_name), cls._get_wal_filename(program_name))

    def _get_absolute_path(self, root_dir: Path, path: Path | str) -> Path | None:
        """Convert paths to relevant absolute paths."""
        # Do not normalize with resolve() to keep symlink paths.
        path = Path(path)

        abs_path = path.absolute()
        if abs_path.is_relative_to(root_dir):
            # absolute path under the root, return the absolute path
            return abs_path

        if not path.is_absolute():
            return (root_dir / path).absolute()

        if root_dir.is_relative_to(path):
            # path is above the root dir. use the root dir.
            return root_dir

        # path is outside our jurisdiction.
        if self._config.verbose:
            cprint(
                f"Timestamp outside {root_dir}'s tree, ignored: {path}",
                "white",
                attrs=["dark"],
            )
        return None

    def _config_yaml(self):
        self._YAML = YAML()
        self._YAML.allow_duplicate_keys = True
        self._YAML.indent(offset=2)  # Conform to Prettier
        self._YAML.representer.add_representer(frozenset, SafeRepresenter.represent_set)
        self._YAML.representer.add_representer(Mapping, SafeRepresenter.represent_dict)
        self._YAML.representer.add_representer(
            MappingProxyType, SafeRepresenter.represent_dict
        )

    def __init__(self, config: TreestampsConfig) -> None:
        """Initialize instance variables."""
        # config
        self._config = config

        # init variables
        # Do not normalize root_dir because symlinks behave weird.
        root_dir = self.get_dir(self._config.path).absolute()
        self.root_dir = root_dir
        self._config_yaml()
        self._filename = self._get_filename(self._config.program_name)
        self._wal_filename = self._get_wal_filename(self._config.program_name)
        self._dump_path = self.root_dir / self._filename
        self._wal_path = self.root_dir / self._wal_filename
        self._wal: TextIO | None = None
        self._consumed_paths: set[Path] = set()
        self._timestamps: dict[Path, float] = {}
